/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2013, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.h
    \brief    prediction class (header)
*/

#ifndef __TCOMPREDICTION__
#define __TCOMPREDICTION__


// Include files
#include "TComPic.h"
#include "TComMotionInfo.h"
#include "TComPattern.h"
#include "TComTrQuant.h"
#include "TComInterpolationFilter.h"
#include "TComWeightPrediction.h"

#if SVC_UPSAMPLING
#include "TComDataCU.h"
#include "TComUpsampleFilter.h"
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
#include "TComMemCalc.h"
#endif //SHVC_COMPLEXITY_ASSESSMENT
//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// prediction class
class TComPrediction : public TComWeightPrediction
{
protected:
  Int*      m_piYuvExt;
  Int       m_iYuvExtStride;
  Int       m_iYuvExtHeight;
  
  TComYuv   m_acYuvPred[2];
  TComYuv   m_cYuvPredTemp;
  TComYuv m_filteredBlock[4][4];
  TComYuv m_filteredBlockTmp[4];
  
  TComInterpolationFilter m_if;
  
#if SVC_UPSAMPLING
  TComUpsampleFilter      m_cUsf;
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalc[ NUM_MEMORY_ARCHITECTURES ];
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalcBlock[ NUM_MEMORY_ARCHITECTURES ];
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalcPic[ NUM_MEMORY_ARCHITECTURES ];
#if SLD_COMPLEXITY_ASSESSMENT
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalcILMCBlock[ NUM_MEMORY_ARCHITECTURES ];
#endif
  TComFrameMemoryAccessCalculator* m_apcMemAccessCalcRec[ NUM_MEMORY_ARCHITECTURES ];
#endif // SHVC_COMPLEXITY_ASSESSMENT
#if REF_EL_ILMC
  TComYuv *m_pILMCPred[2];       ///< ILMC L0; ILMC L1
#endif
  Pel*   m_pLumaRecBuffer;       ///< array for downsampled reconstructed luma sample 
  Int    m_iLumaRecStride;       ///< stride of #m_pLumaRecBuffer array

  Void xPredIntraAng            (Int bitDepth, Int* pSrc, Int srcStride, Pel*& rpDst, Int dstStride, UInt width, UInt height, UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable, Bool bFilter );
  Void xPredIntraPlanar         ( Int* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height );
  
  // motion compensation functions
  Void xPredInterUni            ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi=false          );
  Void xPredInterBi             ( TComDataCU* pcCU,                          UInt uiPartAddr,               Int iWidth, Int iHeight,                         TComYuv*& rpcYuvPred );
  Void xPredInterLumaBlk  ( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xPredInterChromaBlk( TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi );
  Void xWeightedAverage         ( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst );
  
  Void xGetLLSPrediction ( TComPattern* pcPattern, Int* pSrc0, Int iSrcStride, Pel* pDst0, Int iDstStride, UInt uiWidth, UInt uiHeight, UInt uiExt0 );

  Void xDCPredFiltering( Int* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight );
  Bool xCheckIdenticalMotion    ( TComDataCU* pcCU, UInt PartAddr);

public:
  TComPrediction();
  virtual ~TComPrediction();
  
#if REF_EL_ILMC
  Void    initTempBuff(Bool bILMCPred);
#else
  Void    initTempBuff();
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
#if SLD_COMPLEXITY_ASSESSMENT
  Void initMCMemoryAccessCalculator ( TComPic* pcRefPic, int design_flag, Bool bExt = false);
#else
  Void initMCMemoryAccessCalculator ( TComPic* pcRefPic, int design_flag);
#endif
  Void xCalcMCMemoryAccessLuma      ( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
  Void xCalcMCMemoryAccessChroma    ( TComDataCU* pcCU, Int uiPartAddr, RefPicList eRefPicList, Int iRefIdx, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
  Void xCalcUSMemoryAccessLuma      ( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
  Void xCalcUSMemoryAccessChroma    ( TComDataCU* pcCU, Int uiPartAddr, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV);
#if SLD_COMPLEXITY_ASSESSMENT
  Void xCalcILMCMemoryAccessLuma    ( Int BlkX, Int BlkY, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV );
  Void xCalcILMCMemoryAccessChroma  ( Int BlkX, Int BlkY, TComMv* pcMv, Int iWidth, Int iHeight, Int iFilTapH, Int iFilTapV );
#if REF_BL_RESIDUE
  Void xCalcResiMemoryAccessLuma    ( Int BlkX, Int BlkY, Int iWidth, Int iHeight );
  Void xCalcResiMemoryAccessChroma  ( Int BlkX, Int BlkY, Int iWidth, Int iHeight );
#endif
#endif
  Void xCalcRefRrameWritingMemoryAccess( Int iWidth, Int iHeight, Int iLPelX, Int iTPelY);
  Void xCalcRecWritingMemoryAccess( Int iWidth, Int iHeight, Int iLPelX, Int iTPelY);
#endif //SHVC_COMPLEXITY_ASSESSMENT  
  // inter
  Void motionCompensation         ( TComDataCU*  pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList = REF_PIC_LIST_X, Int iPartIdx = -1 );
  
  // motion vector prediction
  Void getMvPredAMVP              ( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, TComMv& rcMvPred );
  
  // Angular Intra
  Void predIntraLumaAng           ( TComPattern* pcTComPattern, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft );
  Void predIntraChromaAng         ( Int* piSrc, UInt uiDirMode, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Bool bAbove, Bool bLeft );
  
  Pel  predIntraGetPredValDC      ( Int* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, Bool bAbove, Bool bLeft );
  
  Int* getPredicBuf()             { return m_piYuvExt;      }
  Int  getPredicBufWidth()        { return m_iYuvExtStride; }
  Int  getPredicBufHeight()       { return m_iYuvExtHeight; }

#if REF_IDX_FRAMEWORK
#if SVC_UPSAMPLING
#if SCALED_REF_LAYER_OFFSETS
  Void upsampleBasePic( UInt refLayerIdc, TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic, const Window window );
#else
  Void upsampleBasePic( UInt refLayerIdc, TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic );
#endif
#endif
#endif

#if REF_EL_ILMC
  Void predILMCPic             (TComPic *pcELPic, TComPic *pcILRPic, TComPic *pcILMCPic, TComList<TComPic*>* pcListPic);
  Void predILMCBlk             (TComPic *pcELPic, TComPic *pcILRPic, TComPic *pcILMCPic, TComMv *pcBlkPos, TComMv * pcBlkSize, TComList<TComPic*>* pcListPic);
  Void xPredInterLumaBlkByPos  (TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi);
  Void xPredInterChromaBlkByPos(TComPicYuv *refPic, Int iBlkX, Int iBlkY,  TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi);
  Void copyBlkPic              (TComPicYuv *pcPicYuvSrc, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight);
#endif
#if REF_BL_RESIDUE
  Void addBlkResiPic(TComPicYuv *pcPicYuvResi, TComPicYuv *pcPicYuvDst, Int iBlkX, Int iBlkY, Int iBlkWidth, Int iBlkHeight);
#endif
#if ILMC_BLK_DEC
  Void predILMCUniPU           (TComDataCU* pcCU, UInt uiPartAddr, Int iWidth,    Int iHeight,                           TComYuv*& rpcYuvPred, Bool bi);
  Void predILMCUniBlk          (TComDataCU* pcCU, UInt uiPartAddr, Int iBlkWidth, Int iBlkHeight, Int iBlkX,  Int iBlkY, TComYuv*& rpcYuvPred, Bool bi);
#endif
#if SHVC_COMPLEXITY_ASSESSMENT
  Void   initMCMemoryAccessCalculator( const MemCmpParam& cLumaParam, const MemCmpParam& cChromaParam, int design_flag );
  Void   updatePic( int design_flag );
  UInt64 getTotalMCMemoryAccessBytes( Int iIndex, int design_flag );
#endif //SHVC_COMPLEXITY_ASSESSMENT
#if SLD_PIX_COMPLEXITY_ASSESSMENT
  Void xCalcBLTransformPix     (TComDataCU* pcCU, UInt uiPartAddr, Int iWidth,    Int iHeight);
#endif
};

//! \}

#endif // __TCOMPREDICTION__
